/*
 * gui_thermal.c
 *
 *  Created on: 2016/5/19
 *      Author: User
 */

#include "ILI9341.h"
#include "gui_thermal.h"
#include <math.h>
#include "simple_graphics.h"
#include "vip_fr.h"
#include "ILI9341.h"
/////////////////////////////////////
// VIP Frame Reader
#define FRAME_WIDTH  80
#define FRAME_HEIGHT 60
#define FRAME_SIZE   (FRAME_WIDTH*FRAME_HEIGHT*4)
#define FR_FRAME_0 	   (SDRAM_BASE + 0x00000000)
#define FR_FRAME_1     (FR_FRAME_0 + FRAME_SIZE)


//////////////////////////////////////
// Color Bar
//void LCD_Init();
//void LCD_Clear(alt_u16 Color);
//void LCD_DrawPoint(alt_u16 x,alt_u16 y,alt_u16 color );


//#define RED 0xf800
//#define GREEN 0x07e0
//#define BLUE 0x001f

// R(5)G(6)B(5)

// http://www.rapidtables.com/convert/color/hsl-to-rgb.htm

#define COLOR_NUM			512
#define GRAY_BASE 			7600
#define COLOR_BAR_WIDTH		20
#ifndef SUPPORT_FONT
	#define COLOR_BAR_TOP		0
	#define COLOR_BAR_BOTTOM	240
#else
	#define COLOR_BAR_TOP		20//50
	#define COLOR_BAR_BOTTOM	200//220
#endif
#define COLOR_BAR_SIZE      (COLOR_BAR_BOTTOM-COLOR_BAR_TOP)


void HSL2RGB(float H, float S, float L, float *pR, float *pG, float *pB){
	float C, X, m, r,g,b, h;

	h = H / 60.0;
	while( h >= 2.0)
		h -= 2.0;

	C = (1.0 - fabs(2.0*L - 1.0)) * S;
	X = C * (1.0 - fabs( h - 1.0) );
	m = L - C/2.0;
//printf("C=%.3f,X=%.3f,m=%.3f, H=%.3f\r\n", C, X, m, H);
	if (H >= 0.0 && H < 60.0){
		r = C;
		g = X;
		b = 0.0;
	}else if (H >= 60.0 && H < 120.0){
			r = X;
			g = C;
			b = 0.0;
	}else if (H >= 120.0 && H < 180.0){
			r = 0.0;
			g = C;
			b = X;
	}else if (H >= 180.0 && H < 240.0){
			r = 0.0;
			g = X;
			b = C;
	}else if (H >= 240.0 && H < 300.0){
			r = X;
			g = 0.0;
			b = C;
	}else{// if (H >= 300.0 && H < 360.0){
			r = C;
			g = 0;
			b = X;
	}
	*pR = (r+m)*255.0;
	*pG = (g+m)*255.0;
	*pB = (b+m)*255.0;

	if (*pR < 0.0)
		*pR = 0.0;
	else if (*pR > 255.0)
		*pR = 255.0;

	if (*pG < 0.0)
		*pG = 0.0;
	else if (*pG > 255.0)
		*pG = 255.0;


	if (*pB < 0.0)
		*pB = 0.0;
	else if (*pB > 255.0)
		*pB = 255.0;

}


void DumpThermal(int count){
	int i, value;

	for(i=0;i<count;i++){
		value = IORD_16DIRECT(LAPTON_BASE, i*2) & 0xFFFF; // read data
		printf("%04xh, ", value);
		if (i%10 == 9)
			printf("\r\n");
	}

}

alt_u32 szTable888[COLOR_NUM]; // R(8)G(8)B(8)
alt_u16 szTable565[COLOR_NUM]; // R(5)G(6)B(5)
void BuildTable(){
	int i;
	alt_u16 Value16;
	alt_u32 Value32;

	float S = 0.95;
	float L = 0.5;
	float H;

	for(i=0;i<COLOR_NUM;i++){
		float r,g,b;

		//S = (float)i/(float)COLOR_NUM;
		//L = (float)i/(float)COLOR_NUM;
		H = (float)i/(float)COLOR_NUM * 180.0;   /* 0 --> 180 */
		H += 240.0;  /* 0+240 --> 180+240 */
		if (H >= 360)
			H -= 360;  /* 0+240 --> 180+240-360=60 */
		// H: 300 --> 120 : blue --> yellow
		HSL2RGB(H,S,L,&r,&g,&b);

		Value16 = 0; // R(5)G(6)B(5)
		Value16 = (alt_u16)(b/8.0);
		Value16 |= ((alt_u16)(g/4.0)) << 5;
		Value16 |= ((alt_u16)(r/8.0)) << 11;
		szTable565[i] = Value16;

		Value32 = 0; // R(8)G(8)B(8)
		Value32 = (alt_u16)(b);
		Value32 |= ((alt_u16)(g)) << 8;
		Value32 |= ((alt_u16)(r)) << 16;
		szTable888[i] = Value32;

	//	printf("[%d]=%xh, r=%.1f, g=%.1f, b=%.1f\r\n", i, Value16, r,g,b);


	}

}


#define COLOR_BLACK 0x000000
#define COLOR_WHITE 0xFFFFFF


void DrawThermal(VIP_FRAME_READER *pReader, alt_video_display *pDisplay, bool bVerticalMirror, bool bHorizontalMirror){
	int x, y,xx, yy, i, j, index, new_index;
	char szText[128];
	alt_u32 *pFrame, Value32;


	const int Scale = 4;
	int Gray14 ;
	int Maxi = 0, Mini = 1024*64, Base, Range, TmpColor;
	alt_u16 Value16;
	alt_u16 szFrame_ORG[80*60];
	alt_u16 szFrame[80*60];
	alt_u16 szScaledLine[60*4*4];
	alt_u16 szColorBar[COLOR_BAR_SIZE];

	// read & find max & mini
	for(index=0;index<60*80;index++){
			Gray14 = IORD_16DIRECT(LAPTON_BASE, index*2) & 0x3FFF;

			// find max & mini
			if (Gray14 > Maxi)
				Maxi = Gray14;
			if (Gray14 < Mini)
				Mini = Gray14;


			szFrame_ORG[index] = Gray14;
	} // for index
	Base = Mini;
	Range = Maxi - Mini;


	///////////////////////////////
	// draw on VIP Frame Reader x1
	pFrame = VIPFR_GetDrawFrame(pReader);
	for(y=0;y<60;y++){
		for(x=0;x<80;x++){
			if (x <= 5){
				// draw color bar
				if (x == 5){
					Value32 = 0xFFFFFF; // white
				}else{
					if (bVerticalMirror)
						TmpColor = (COLOR_NUM-1)*y/60;  // scale
					else
						TmpColor = (COLOR_NUM-1)*(60 - y - 1)/60;  // scale
					Value32 = szTable888[TmpColor]; // map
				}
			}else{
				xx = bVerticalMirror?(80-x-1):x;
				yy = bHorizontalMirror?y:(60 - y - 1);
				index = yy*80 + xx;
				Gray14 = szFrame_ORG[index];

				// scale to 0~(COLOR_NUM-1)
				TmpColor = Gray14-Base;
				TmpColor = (COLOR_NUM-1)*TmpColor/Range;  // scale
				Value32 = szTable888[TmpColor]; // map
			}


			*pFrame++ = Value32;
		} // for xx
	} // for yy
	VIPFR_ActiveDrawFrame(pReader);


	///////////////////////////////
	// draw on LT24

	// xfer gray to temperature color, and change x->y, y=SCREEN_HEIGHT-x
	index = 0;
	for(y=0;y<60;y++){
		for(x=0;x<80;x++){
			Gray14 = szFrame_ORG[index];
			index++;

			// scale to 0~(COLOR_NUM-1)
			TmpColor = Gray14-Base;
			TmpColor = (COLOR_NUM-1)*TmpColor/Range;  // sclae

			Value16 = szTable565[TmpColor]; // map

			xx = bVerticalMirror?(60-y-1):y;
			yy = bHorizontalMirror?(80 - x - 1):x;
			new_index = yy*60 + xx;
			szFrame[new_index] = Value16; // 60x80

		} // for x
	} // fox y



	// draw on LT24-screen by x4
	for(yy=6;yy<80;yy++){
		index = yy*60;
		for(xx=0;xx<60;xx++){
			Value16 = szFrame[index];
			index++;
			for(i=0;i<Scale;i++){
				new_index = i*60*Scale + xx*Scale;
				for(j=0;j<Scale;j++){
					szScaledLine[new_index] = Value16;
					new_index++;
				}
			}
		} // for xx
		LCD_DrawBlock(0, yy*Scale, szScaledLine, 60*Scale*Scale);
	} // for yy

	// draw temperature bar
	for(xx=COLOR_BAR_TOP;xx<COLOR_BAR_BOTTOM;xx++){
		index = (int)((float)(xx-COLOR_BAR_TOP)/(float)(COLOR_BAR_BOTTOM-COLOR_BAR_TOP)*(float)COLOR_NUM);
		if (index >= COLOR_NUM)
			index = COLOR_NUM - 1;
		//index = COLOR_NUM - index - 1;
		szColorBar[xx-COLOR_BAR_TOP] = szTable565[index];
	}
	for(yy=0;yy<COLOR_BAR_WIDTH;yy++)
		LCD_DrawBlock(COLOR_BAR_TOP, yy, szColorBar, COLOR_BAR_SIZE);


#ifdef SUPPORT_FONT

	// center point temperature
	sprintf(szText,"%d", szFrame_ORG[80*29+39]-Mini);
	vid_print_string_alpha(0,SCREEN_HEIGHT-20, COLOR_WHITE, COLOR_BLACK, tahomabold_20, pDisplay, szText);


	// max temperature
	sprintf(szText,"%d", Range);
//	vid_print_string_alpha(0,SCREEN_HEIGHT-20, COLOR_WHITE, COLOR_BLACK, tahomabold_20, pDisplay, szText);
	vid_print_string_alpha(COLOR_BAR_BOTTOM+2, 0,COLOR_WHITE, COLOR_BLACK, tahomabold_20, pDisplay, szText);

	// mini temperature
	sprintf(szText,"%d", 0);
//	vid_print_string_alpha(COLOR_BAR_BOTTOM+2, 0, COLOR_WHITE, COLOR_BLACK, tahomabold_20, pDisplay, szText);
	vid_print_string_alpha(0, 0, COLOR_WHITE, COLOR_BLACK, tahomabold_20, pDisplay, szText);

	vid_draw_line(SCREEN_WIDTH/2-10, SCREEN_HEIGHT/2, SCREEN_WIDTH/2+10, SCREEN_HEIGHT/2, 1, COLOR_WHITE, pDisplay);
	vid_draw_line(SCREEN_WIDTH/2, SCREEN_HEIGHT/2-10, SCREEN_WIDTH/2, SCREEN_HEIGHT/2+10, 1, COLOR_WHITE, pDisplay);
#endif //SUPPORT_FONT

	printf("Maxi=%d, Mini=%d\r\n", Maxi, Mini);
//	printf("Maxi=%d, Mini=%d,bVerticalMirror=%d\r\n", Maxi, Mini, bVerticalMirror);

}





// Pixel data is 14 bit, with some automatic flat-field correction. Room temperature is around 0x2000.
void GUI_Thermal(alt_video_display *pDisplay, TOUCH_HANDLE *pTouch){
	bool bVerticalMirror, bHorizontalMirror;
	void *pFrameBuffer0;
	void *pFrameBuffer1;
	void *pFrameBuffer;
	VIP_FRAME_READER *pReader;

	// LT24 INIT
	LCD_Init();
	LCD_Clear(0x000000);

	////////////////////////////
	// VIP FrameReader

	// Init
	/*pFrameBuffer0 = (void *)malloc(FRAME_SIZE);
	if (pFrameBuffer0 == 0)
		printf("failed to alloc - pFrameBuffer0 \r\n");
	pFrameBuffer1 = (void *)malloc(FRAME_SIZE);
	if (pFrameBuffer1 == 0)
		printf("failed to alloc - pFrameBuffer1 \r\n");*/
    pReader =  VIPFR_Init(ALT_VIP_VFR_0_BASE, (void *)FR_FRAME_0, (void *)FR_FRAME_1, FRAME_WIDTH, FRAME_HEIGHT);
    VIPFR_Go(pReader, TRUE);

    // draw background - white
    pFrameBuffer = VIPFR_GetDrawFrame(pReader);
    memset((void *)pFrameBuffer, 0xFF, FRAME_SIZE);
    VIPFR_ActiveDrawFrame(pReader);



	// build table
	BuildTable();
	DumpThermal(40);

	while(1){
		bVerticalMirror = IORD(SWITCH_BASE, 0x00) & 0x01;
		bHorizontalMirror = IORD(SWITCH_BASE, 0x00) & 0x02;
		DrawThermal(pReader, pDisplay, bVerticalMirror, bHorizontalMirror);
		//printf(".");
	}

}

